/*
* This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/bonsai
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
* 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
* 				Marian Pohling, Lukas Twardon
*/
package de.unibi.airobots.bonsai.actuators;

import java.io.IOException;

import de.unibi.airobots.btl.data.CameraPositionData;
import de.unibi.airobots.btl.data.CameraTargetData;
import de.unibi.airobots.bonsai.core.Actuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.sensors.data.PersonList;

/**
 * New interface to control a pan-tilt camera.
 * 
 * @author lziegler
 * @author ttoenige
 */
public interface CameraActuator extends Actuator {

    /**
     * Possible camera speeds.
     * 
     * @author lziegler
     */
    public enum CameraSpeed {
        SLOW, MEDIUM, FAST
    };

    /**
     * Set a target for the camera to look at.
     * 
     * @param target
     *            The target to look at.
     * @throws IOException
     *             Is thrown if some communication error occure.
     */
    void setTarget(CameraTargetData target) throws IOException;

    /**
     * Getter for the height of the camera above the ground.
     * 
     * @return height in mm.
     * @throws IOException
     *             Is thrown if some communication error occure.
     */
    long getCameraHeight() throws IOException;

    /**
     * Set a pan/tilt/zoom position of the camera.
     * 
     * @param position
     *            The position to move to.
     * @throws IOException
     *             Is thrown if some communication error occure.
     */
    void setPosition(CameraPositionData position) throws IOException;

    /**
     * Set the speed pan and tilt movements are done.
     * 
     * @param speed
     *            The speed used for pan and tilt movements.
     * 
     * @throws IOException
     *             Is thrown if some communication error occure.
     */
    void setSpeed(CameraSpeed speed) throws IOException;

    /**
     * Getter for the current pan/tilt/zoom values of the camera.
     * 
     * @return The current position.
     * @throws IOException
     *             Is thrown if some communication error occure.
     */
    CameraPositionData getPosition() throws IOException;

    /**
     * Focus a special person.
     * 
     * @param personID
     *            person to focus
     */
    void focus(Sensor<PersonList> personSensor, int personId) throws IOException;

    /**
     * stop focusing a special person
     */
    void unfocus() throws IOException;

    /**
     * start looking at persons
     */
    void startCamAction(Sensor<PersonList> personSensor) throws IOException;

    /**
     * stop looking at persons
     */
    void stopCamAction() throws IOException;
}
