/*
* This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/bonsai
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
* 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
* 				Marian Pohling, Lukas Twardon
*/
package de.unibi.airobots.bonsai.examples;

import java.awt.Component;
import java.awt.Container;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.net.URISyntaxException;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextArea;
import javax.swing.JToggleButton;

import net.sf.xcf.InitializeException;
import net.sf.xcf.Publisher;
import net.sf.xcf.Subscriber;
import net.sf.xcf.XcfManager;
import net.sf.xcf.event.PublishEvent;
import net.sf.xcf.event.PublishEventListener;
import net.sf.xcf.event.QueueAdapter;
import net.sf.xcf.event.XcfEvent;
import net.sf.xcf.event.system.AdditionEvent;
import net.sf.xcf.event.system.DeletionEvent;
import net.sf.xcf.event.system.SystemEvent;
import net.sf.xcf.event.system.SystemEventListener;
import net.sf.xcf.examples.SystemEventPrinter;
import net.sf.xcf.ice.IceSubscriber;
import net.sf.xcf.naming.NameNotFoundException;
import net.sf.xcf.util.SynchronizedQueue;
import net.sf.xcf.xml.XPath;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.actuators.EnterArenaAndLocalizeActuator;
import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.actuators.SpeechActuator;
import de.unibi.airobots.bonsai.actuators.SwissRangerControl;
import de.unibi.airobots.bonsai.core.BonsaiManager;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.core.BonsaiManager.ConfigurationException;
import de.unibi.airobots.bonsai.util.CoordinateSystemConverter;
import de.unibi.airobots.bonsai.util.slam.SlamTools;
import de.unibi.airobots.btl.data.BinarySlamMap;
import de.unibi.airobots.btl.data.GlobalPlan;
import de.unibi.airobots.btl.data.NavigationGoalData;
import de.unibi.airobots.btl.data.PositionData;

/**
 * This Class contains an example showing a way to control Biron with a "remote"
 * (actually your remote is a GUI). It's containing basic operations like making
 * him move, turn or say something.
 */
public class BasicRemote extends JPanel implements ActionListener {

    private static final long serialVersionUID = 72835335739325107L;

    /**
     * Startup main-method.
     * 
     * @param args
     *            none
     */
    public static void main(String[] args) {
        try {
            new BasicRemote();
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    /**
     * Initialization of ...
     */

    /**
     * ... swing-components
     */
    private JTextArea infos;
    private JTextArea targets;
    private JTextArea speechtxt;

    /**
     * ... actuators and sensors
     */

    private NavigationActuator navigationActuator;
    private Sensor<BinarySlamMap> slamSensor;
    private SwissRangerControl srControl;
    private SpeechActuator speechActuator;
    private Sensor<GlobalPlan> planSensor;
    private EnterArenaAndLocalizeActuator enterArenaActuator;

    /**
     * ... variables for coordinates
     */
    private Double xCoordinate = 0.0;
    private Double yCoordinate = 0.0;
    private Double yawAngle = 0.0;
    private Double xGoalCoordinate = 0.0;
    private Double yGoalCoordinate = 0.0;
    private Double goalYawAngle = 0.0;
    /**
     * ... the logger
     */
    private static final Logger LOGGER = Logger.getLogger(BasicRemote.class);
    private static final double STEP_FORWARD = 2.0;
    private JToggleButton srActivate;
    private XcfManager manager;
    private Subscriber sub;
    private Publisher rr;
    private XcfManager m;
    private SynchronizedQueue<PublishEvent> queue;
    private XcfEvent result;

    /**
     * Constructor.
     * 
     * @throws URISyntaxException
     * @throws ConfigurationException
     * @throws IllegalStateException
     */
    public BasicRemote() throws IllegalStateException, ConfigurationException,
            URISyntaxException {

        /**
         * Setting up the actuators and sensors
         */

        /*
         * While programming Biron one can send information to the logger. First
         * of all you need to initialize the logger and configure it.
         */
        BasicConfigurator.configure();

        /*
         * As there may be different situations in which something is logged
         * there are different logging levels. For example there is:
         * .info(message) .debug(message) .fatal(message) .error(message)
         * .warn(message) Moreover one can configure the logger and set a level
         * so that it will only show logged messages in this level.
         */
        LOGGER.info("setting up...");

        /*
         * Each actuator and sensor has its own manager, organizing each
         * actuators/sensors availability status. First one have to create a new
         * manager (if there is none yet). After that it is possible to get the
         * actuator/sensor itself but requesting it from the manager:
         * 
         * managerOfYourChoice.getActuator(); (or .getSensor();)
         * 
         * As the manager itself is not needed for the remaining interaction
         * with Biron there is also an alternative way to get the
         * actuator/sensor: The BonsaiFactory. It holds methods to create
         * Managers and allows you to create a manager and get get the reference
         * to the actuator/sensor immediately:
         * 
         * BonsaiFactory.getXManager().getActuator(); (X stands for any
         * actuator/sensor)
         */
        LOGGER.info("... NavigationManager ...");

        BonsaiManager.getInstance().configure(
                "/vol/robocup/trunk/etc/bonsai_config/basicremote.xml");

        navigationActuator =
                BonsaiManager.getInstance().createActuator(
                        "NavigationActuator", NavigationActuator.class);
        slamSensor =
                BonsaiManager.getInstance().createSensor("SlamSensor",
                        BinarySlamMap.class);
        // TODO will generate a NullPointerException if used
        // srControl = BonsaiManager.getInstance().createActuator(
        // "SwissrangerControl", SwissRangerControl.class);
        speechActuator =
                BonsaiManager.getInstance().createActuator("SpeechActuator",
                        SpeechActuator.class);
        planSensor =
                BonsaiManager.getInstance().createSensor("GlobalPlanSensor",
                        GlobalPlan.class);
        Sensor<PositionData> posSensor =
                BonsaiManager.getInstance().createSensor("PositionSensor",
                        PositionData.class);

        Sensor<BinarySlamMap> slamSensor2 =
                BonsaiManager.getInstance().createSensor("SlamSensor",
                        BinarySlamMap.class);
        SlamTools.showMapPanelWindow(slamSensor2, posSensor,
                navigationActuator, planSensor);

        enterArenaActuator =
                BonsaiManager.getInstance().createActuator(
                        "EnterArenaAndLocalizeActuator",
                        EnterArenaAndLocalizeActuator.class);

        /**
         * Setting-up the Gui
         */
        this.guiSetup();
        System.out.println("GUI READY");
    }

    /**
     * Containing everything concerning the GUI.
     */
    private void guiSetup() {
        JFrame gui = new JFrame("BasicRemote for Biron");
        gui.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        gui.setLocation(100, 100);

        JButton buttonUpdate = new JButton("UPDATE");
        buttonUpdate.addActionListener(this);

        JButton buttonUp = new JButton("Forward");
        buttonUp.addActionListener(this);

        JButton buttonDown = new JButton("Backward");
        buttonDown.addActionListener(this);

        JButton buttonLeft = new JButton("Left");
        buttonLeft.addActionListener(this);

        JButton buttonRight = new JButton("Right");
        buttonRight.addActionListener(this);

        JButton buttonStop = new JButton("STOP");
        buttonStop.addActionListener(this);

        JButton bTurnLeft = new JButton("45° Turn Left");
        bTurnLeft.addActionListener(this);

        JButton bTurnRight = new JButton("45° Turn Right");
        bTurnRight.addActionListener(this);

        JButton bTurnLeft10 = new JButton("20° Turn Left");
        bTurnLeft10.addActionListener(this);

        JButton bTurnRight10 = new JButton("20° Turn Right");
        bTurnRight10.addActionListener(this);

        JButton bsay = new JButton("SAY");
        bsay.addActionListener(this);

        JButton bsay1 = new JButton("hello");
        bsay1.addActionListener(this);

        JButton bsay2 = new JButton("EPFL");
        bsay2.addActionListener(this);

        JButton bsay3 = new JButton("bye");
        bsay3.addActionListener(this);

        JButton bsay4 = new JButton("like robots");
        bsay4.addActionListener(this);

        JButton bsay5 = new JButton("thank you");
        bsay5.addActionListener(this);

        JButton camerasensor = new JButton("Camera-Sensor");
        camerasensor.addActionListener(this);

        JButton localize = new JButton("Localize");
        localize.addActionListener(this);

        JButton poseActuator = new JButton("Arm Control");
        poseActuator.addActionListener(this);

        srActivate = new JToggleButton("SwissRanger", false);
        srActivate.addActionListener(this);

        speechtxt = new JTextArea("Enter text here to say", 1, 5);
        infos = new JTextArea("", 3, 5);
        infos.setEditable(false);
        infos.setText(getInfos());

        targets = new JTextArea();
        targets.setEditable(false);
        targets.setColumns(2);

        update();
        setTexts();

        Container pane = gui.getContentPane();
        GridBagLayout gbl = new GridBagLayout();
        pane.setLayout(gbl);

        // cont, gbl, comp, x, y, width, height, weightx, weighty
        // CHECKSTYLE:OFF magic numbers are ok here
        addComponent(pane, gbl, infos, 1, 0, 1, 1, 1.0, 0);
        addComponent(pane, gbl, targets, 3, 0, 1, 1, 1.0, 0);
        addComponent(pane, gbl, new JLabel(" "), 0, 1, 4, 1, 0, 0);
        addComponent(pane, gbl, buttonUp, 2, 2, 1, 1, 0, 0);
        addComponent(pane, gbl, buttonLeft, 1, 3, 1, 1, 0, 0);
        addComponent(pane, gbl, buttonRight, 3, 3, 1, 1, 0, 0);
        addComponent(pane, gbl, buttonDown, 2, 4, 1, 1, 0, 0);
        addComponent(pane, gbl, buttonStop, 2, 3, 1, 1, 0, 0);
        addComponent(pane, gbl, new JLabel(" "), 0, 5, 4, 1, 0, 0);
        addComponent(pane, gbl, bTurnLeft10, 1, 6, 1, 1, 0, 0);
        addComponent(pane, gbl, bTurnLeft, 1, 7, 1, 1, 0, 0);
        addComponent(pane, gbl, bTurnRight10, 3, 6, 1, 1, 0, 0);
        addComponent(pane, gbl, bTurnRight, 3, 7, 1, 1, 0, 0);
        addComponent(pane, gbl, buttonUpdate, 1, 8, 3, 1, 0, 0);
        addComponent(pane, gbl, new JLabel(" "), 0, 9, 4, 1, 0, 0);
        addComponent(pane, gbl, bsay1, 1, 10, 1, 1, 1.0, 0);
        addComponent(pane, gbl, bsay2, 2, 10, 1, 1, 1.0, 0);
        addComponent(pane, gbl, bsay3, 3, 11, 1, 1, 1.0, 0);
        addComponent(pane, gbl, bsay4, 1, 11, 1, 1, 1.0, 0);
        addComponent(pane, gbl, bsay5, 2, 11, 1, 1, 1.0, 0);
        addComponent(pane, gbl, bsay, 3, 12, 1, 1, 1.0, 0);
        addComponent(pane, gbl, speechtxt, 1, 12, 2, 1, 1.0, 0);
        addComponent(pane, gbl, camerasensor, 2, 1, 1, 1, 0, 0);
        addComponent(pane, gbl, localize, 2, 0, 1, 1, 0, 0);
        addComponent(pane, gbl, new JLabel(" "), 0, 13, 4, 1, 0, 0);
        addComponent(pane, gbl, poseActuator, 2, 14, 1, 1, 0, 0);
        addComponent(pane, gbl, srActivate, 3, 14, 1, 1, 0, 0);
        // CHECKSTYLE:ON

        gui.pack();
        gui.setVisible(true);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void actionPerformed(ActionEvent e) {

        try {

            if (e.getActionCommand().equals("UPDATE")) {
                update();
            }

            if (e.getActionCommand().equals("SwissRanger")) {
                if (srActivate.isSelected()) {
                    LOGGER.info("Activating SwissRanger");
                    srControl.activateVirtualLaser();
                } else {
                    LOGGER.info("Deactivating SwissRanger");
                    srControl.deactivateVirtualLaser();
                }
            }

            /*
             * The following else ifs contain basic functions to control Biron.
             * As the interaction with Biron is often the same only one example
             * is explained.
             */

            else if (e.getActionCommand().equals("Forward")) {
                update();
                /*
                 * First of all you have to create a goal for Biron. This is
                 * done by creating a NavigationGoalData object by using the
                 * CoordinateSystemConverter like below. It needs the current
                 * global position (OdometryData) of the robot obtained from the
                 * OdometrySensor, the angle and distance of the new goal. In
                 * addition to that it is possible to give a coordinate and yaw
                 * tolerance but by default these are set to zero.
                 * 
                 * Then you have to set the goal by delivering it to the
                 * navigation actuator via .setGoal(goal).
                 * 
                 * In case the navigation actuator is unavailable there is a
                 * try-catch surrounding the instructions.
                 */
                NavigationGoalData g =
                        CoordinateSystemConverter.polar2NavigationGoalData(
                                slamSensor.readLast(), 0.0, 1.0);
                g.setYawTolerance(0.1 * Math.PI);
                setGoal(g);
                navigationActuator.setGoal(g);

            } else if (e.getActionCommand().equals("Backward")) {
                update();
                NavigationGoalData g =
                        CoordinateSystemConverter.polar2NavigationGoalData(
                                slamSensor.readLast(), Math.PI, 1.0);
                g.setYawTolerance(0.1 * Math.PI);
                setGoal(g);
                navigationActuator.setGoal(g);
            } else if (e.getActionCommand().equals("Left")) {
                update();
                NavigationGoalData g =
                        CoordinateSystemConverter.polar2NavigationGoalData(
                                slamSensor.readLast(), Math.PI / 2, 1.0);
                g.setYawTolerance(0.1 * Math.PI);
                setGoal(g);
                navigationActuator.setGoal(g);
            }

            else if (e.getActionCommand().equals("Right")) {
                update();
                NavigationGoalData g =
                        CoordinateSystemConverter.polar2NavigationGoalData(
                                slamSensor.readLast(), -Math.PI / 2, 1.0);
                g.setYawTolerance(0.1 * Math.PI);
                setGoal(g);
                navigationActuator.setGoal(g);
            }

            else if (e.getActionCommand().equals("STOP")) {
                update();
                navigationActuator.manualStop();
            } else if (e.getActionCommand().equals("45° Turn Left")) {
                update();
                NavigationGoalData g =
                        CoordinateSystemConverter.polar2NavigationGoalData(
                                slamSensor.readLast(), Math.PI / 4, 0.0);
                g.setYawTolerance(0.1 * Math.PI);
                setGoal(g);
                navigationActuator.setGoal(g);
            } else if (e.getActionCommand().equals("45° Turn Right")) {
                update();
                NavigationGoalData g =
                        CoordinateSystemConverter.polar2NavigationGoalData(
                                slamSensor.readLast(), -Math.PI / 4, 0.0);
                g.setYawTolerance(0.1 * Math.PI);
                setGoal(g);
                navigationActuator.setGoal(g);
            } else if (e.getActionCommand().equals("22.5° Turn Left")) {
                update();
                NavigationGoalData g =
                        CoordinateSystemConverter.polar2NavigationGoalData(
                                slamSensor.readLast(), Math.PI / 8, 0.0);
                g.setYawTolerance(0.1 * Math.PI);
                setGoal(g);
                navigationActuator.setGoal(g);
            } else if (e.getActionCommand().equals("22.5° Turn Right")) {
                update();
                NavigationGoalData g =
                        CoordinateSystemConverter.polar2NavigationGoalData(
                                slamSensor.readLast(), -Math.PI / 8, 0.0);
                g.setYawTolerance(0.1 * Math.PI);
                setGoal(g);
                navigationActuator.setGoal(g);
            } else if (e.getActionCommand().equals("SAY")) {
                System.err.println(speechtxt.getText());
                speechActuator.say(speechtxt.getText(), true);
            } else if (e.getActionCommand().equals("hello")) {
                speechActuator
                        .say("Hello, would you like to answer some questions?");
            } else if (e.getActionCommand().equals("EPFL")) {
                speechActuator.say("Are you a student here at E P F L?");
            } else if (e.getActionCommand().equals("bye")) {
                speechActuator.say("Okay, have a nice evening!");
            } else if (e.getActionCommand().equals("like robots")) {
                speechActuator.say("Do you like robots?");
            } else if (e.getActionCommand().equals("thank you")) {
                speechActuator.say("Thank you, have a nice evening.");
            } else if (e.getActionCommand().equals("Camera-Sensor")) {

                CameraSensorTest cst = new CameraSensorTest();
                Thread thread = new Thread(cst);
                thread.start();
            } else if (e.getActionCommand().equals("Arm Control")) {
                ArmControlRemote acr = new ArmControlRemote();
            } else if (e.getActionCommand().equals("Localize")) {
                enterArenaActuator.enterArenaAndLocalize(STEP_FORWARD, 0, 0, 2);
                //enterArenaActuator.waitUntilLocalizationFinished();
            }

        } catch (Exception ex) {
            LOGGER.error("Error performing action", ex);
        }
    }

    /**
     * A useful function to add all components into the GridBagLayout.
     */
    private static void addComponent(Container cont, GridBagLayout gbl,
            Component c, int x, int y, int width, int height, double weightx,
            double weighty) {
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.fill = GridBagConstraints.BOTH;
        gbc.gridx = x;
        gbc.gridy = y;
        gbc.gridwidth = width;
        gbc.gridheight = height;
        gbc.weightx = weightx;
        gbc.weighty = weighty;
        gbl.setConstraints(c, gbc);
        cont.add(c);
    }

    private void setTexts() {
        infos.setText(getInfos());
        targets.setText(getTargets());
    }

    private void setGoal(NavigationGoalData g) {
        xGoalCoordinate = g.getX();
        yGoalCoordinate = g.getY();
        goalYawAngle = g.getYaw();
    }

    private String getTargets() {

        if (xGoalCoordinate.toString().length() > 4
                && yGoalCoordinate.toString().length() > 4
                && goalYawAngle.toString().length() > 4) {
            return new String("Goal:\nX: "
                    + xGoalCoordinate.toString().substring(0, 5) + "\nY: "
                    + yGoalCoordinate.toString().substring(0, 5) + "\nYaw: "
                    + goalYawAngle.toString().substring(0, 5));
        } else {
            return new String("Goal:\nX: " + xGoalCoordinate + "\nY: "
                    + yGoalCoordinate + "\nYaw: " + goalYawAngle);
        }
    }

    private String getInfos() {

        if (xCoordinate.toString().length() > 4
                && yCoordinate.toString().length() > 4
                && yawAngle.toString().length() > 4) {
            return new String("Cur Pos:\nX: "
                    + xCoordinate.toString().substring(0, 5) + "\nY: "
                    + yCoordinate.toString().substring(0, 5) + "\nYaw: "
                    + yawAngle.toString().substring(0, 5));
        } else {
            return new String("Cur Pos:\nX: " + xCoordinate + "\nY: "
                    + yCoordinate + "\nYaw: " + yawAngle);
        }
    }

    /**
     * An update function for the GUI. If anything happens the displayed
     * coordinates are updated.
     */
    private void update() {
        System.out.println("Update");
        try {
            System.out.println("get position");
            BinarySlamMap slamData = slamSensor.readLast();
            xCoordinate = slamData.getX();
            yCoordinate = slamData.getY();
            yawAngle = slamData.getYaw();
            setTexts();

        } catch (IOException e) {

            e.printStackTrace();
        } catch (InterruptedException e) {

            e.printStackTrace();
        }
        System.out.println("update done!");
    }
}
