/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 * 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 * 				Marian Pohling, Lukas Twardon
 */
package de.unibi.airobots.bonsai.skills.state;

import java.io.IOException;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.actuators.SpeechActuator;
import de.unibi.airobots.bonsai.core.BonsaiManager;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.core.SensorListener;
import de.unibi.airobots.bonsai.skills.SkillStateMachine;
import de.unibi.airobots.btl.data.LaserData;
import de.unibi.airobots.btl.data.SpeedData;

/**
 * Implementation of a simple follow person state.
 * 
 * @author nkoester
 */
public class WaitForOpenDoor extends AbstractState implements
        SensorListener<LaserData> {

    private Logger logger = Logger.getLogger(this.getClass());
    private double[] oldValues = null;
    private boolean isOpen = false;
    private final int scanCount = 361;

    // Managers and actuators
    private SpeechActuator speechActuator;
    private Sensor<SpeedData> speedSensor;
    private Sensor<LaserData> laserSensor;

    public WaitForOpenDoor(SkillStateMachine aStateMachine) {
        super(aStateMachine);
    }

    public void getRequiredSensorsAndActuators() {
        this.speechActuator = stateMachine.getManagedActuator("SpeechActuator", SpeechActuator.class);
        this.laserSensor = stateMachine.getManagedSensor("LaserSensor", LaserData.class);

        this.speedSensor =
                BonsaiManager.getInstance().createSensor("SpeedSensor",
                        SpeedData.class);
    }

    @Override
    public boolean init() {
        System.out.println("\t" + this.getClass().getSimpleName().toString()
                + " init");

        logger.debug("Wait for open door started...");

        this.laserSensor.addSensorListener(this);
        try {
            speechActuator.sayAccentuated("Knock knock, please open the door.");
        } catch (IOException e) {
            e.printStackTrace();
        }

        return true;
    }

    @Override
    public boolean execute() {
        System.out.println("\t" + this.getClass().getSimpleName().toString()
                + " execute");

        while (!isOpen) {
            try {
                Thread.sleep(500);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }

        return true;
    }

    @Override
    public boolean end() {

        System.out.println("\t" + this.getClass().getSimpleName().toString()
                + " end");

        this.laserSensor.removeSensorListener(this);

        return true;
    }

    @Override
    public void newDataAvailable(LaserData newData) {
        try {
            if (this.speedSensor.readLast().getSpeedTranslation() == 0) {
                double[] scanValues = newData.getScanValues();

                if (oldValues == null) {
                    oldValues = scanValues;
                } else {
                    double error = 0;
                    for (int i = scanCount / 2 - 5; i < scanCount / 2 + 5; i++) {
                        error = +oldValues[i] - scanValues[i];
                    }
                    error = error / 10;
                    if (error == 0)
                        return;

                    if (error < 0 && Math.abs(error) > 0.15) {
                        logger.debug("door open - error: " + error);
                        isOpen = true;
                    }

                    if (error > 0 && Math.abs(error) > 0.15) {
                        logger.debug("door closed - error: " + error);
                    }
                    oldValues = scanValues;
                }
            }
        } catch (Exception e) {
            System.out.println("LOLOL POKEMON EXEPTION CATCHING!!!");
        }
    }
}
