/*
 * Copyright (C) 2011 malinke, this code is free software.
 */
package de.unibi.airobots.bonsai.strategies;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.util.CoordinateSystemConverter;
import de.unibi.airobots.bonsai.util.helper.GoalReachedWaiter;
import de.unibi.airobots.bonsai.util.helper.GoalReachedWaiter.ReachStatus;
import de.unibi.airobots.btl.data.GlobalPlannerStateChange;
import de.unibi.airobots.btl.data.NavigationGoalData;
import de.unibi.airobots.btl.data.NavigationGoalData.GoalType;
import de.unibi.airobots.btl.data.PositionData;
import de.unibi.airobots.btl.data.SpeedData;
import java.io.IOException;
import org.apache.log4j.Logger;

/**
 * Tries to reach navigation goals.
 * @author malinke
 */
public class NavigationStrategy {

    /** Default value for navigation goals. */
    public static final double DEFAULT_COORDINATE_TOLERANCE = 0.25;
    /** Default value for navigation goals. */
    public static final double DEFAULT_YAW_TOLERANCE = 2 * Math.PI;
    /** Milliseconds to wait for reaching a goal. */
    public static final long TIMEOUT = 150000;
    private static final Logger LOGGER =
            Logger.getLogger(NavigationStrategy.class);

    /** Thrown if a goal could not be reached. */
    public static final class NotReachedException extends Exception {

        private final ReachStatus status;

        private NotReachedException(ReachStatus status) {
            super("Navigation goal could not be reached: " + status);
            this.status = status;
        }

        /**
         * Indicates why a goal could not be reached.
         * @return actual status
         */
        public ReachStatus getStatus() {
            return status;
        }
    }
    private final Sensor<PositionData> positionSensor;
    private final NavigationActuator navigationActuator;
    private final GoalReachedWaiter waiter;
    private double coordinateTolerance = DEFAULT_COORDINATE_TOLERANCE;
    private double yawTolerance = DEFAULT_YAW_TOLERANCE;

    /**
     * Initializes this strategy.
     * @param navigationActuator to set goals and ask for reaching
     * @param positionSensor to check reaching
     * @param speedSensor to check reaching
     * @param gplanChangeSensor to check reaching
     */
    public NavigationStrategy(NavigationActuator navigationActuator,
            Sensor<PositionData> positionSensor, Sensor<SpeedData> speedSensor,
            Sensor<GlobalPlannerStateChange> gplanChangeSensor) {
        this.positionSensor = positionSensor;
        this.navigationActuator = navigationActuator;
        this.waiter = new GoalReachedWaiter(positionSensor,
                speedSensor, gplanChangeSensor, navigationActuator);
    }

    /**
     * Tries to reach the given coordinates globally and, on failure, locally.
     * @param target position to reach
     * @throws de.unibi.airobots.robocup.demochallenge2011.NavigationStrategy.NotReachedException
     * if the goal could not be reached, contains the ReachStatus
     * @throws IOException if navigating fails
     */
    public synchronized void driveTo(PositionData target)
            throws NotReachedException, IOException {
        NavigationGoalData goal = new NavigationGoalData("NavigationStrategy",
                target.getX(), target.getY(), target.getYaw(),
                coordinateTolerance, yawTolerance, GoalType.GLOBAL);
        ReachStatus reached = driveTo(goal);
        if (reached.equals(ReachStatus.NOT_REACHED_BLOCKED)) {
            LOGGER.info("Global plan didn't find a path. Trying locally...");
            goal.setType(GoalType.LOCAL);
            reached = driveTo(goal);
        }
        if (!reached.equals(ReachStatus.REACHED)) {
            throw new NotReachedException(reached);
        }
    }

    /**
     * Tries to reach the given goal and returns the resulting status.
     * @param target position to reach
     * @return indicates whether the goal could be reached or not (and why)
     * @throws IOException if navigating fails
     */
    public synchronized ReachStatus driveTo(NavigationGoalData target)
            throws IOException {
        LOGGER.debug("Drive to position " + target);
        waiter.reset();
        navigationActuator.setGoal(target);
        LOGGER.debug("Waiting " + TIMEOUT + "ms for reaching goal...");
        ReachStatus reached = waiter.waitForReachingGoal(target, TIMEOUT);
        LOGGER.debug("Status: " + reached + ". Stopping here.");
        navigationActuator.manualStop();
        return reached;
    }

    /**
     * Turns the robot.
     * @param angle to turn
     * @throws IOException if navigating fails
     * @throws de.unibi.airobots.bonsai.strategies.NavigationStrategy.NotReachedException
     * if the goal could not be reached, contains the ReachStatus
     */
    public void turn(double angle) throws IOException, NotReachedException {
        LOGGER.info("Turning around " + angle + " radiants.");
        NavigationGoalData target;
        try {
            target = CoordinateSystemConverter.polar2NavigationGoalData(
                    positionSensor.readLast(), angle, 0);
        } catch (InterruptedException ex) {
            throw new NotReachedException(ReachStatus.NOT_REACHED_INTERRUPTED);
        }
        target.setYawTolerance(Math.PI / 8.0);
        ReachStatus reached = driveTo(target);
        LOGGER.info("Turning around finished: reached = " + reached);
        if (!reached.equals(ReachStatus.REACHED)) {
            throw new NotReachedException(reached);
        }
    }
}
