/*
* This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/bonsai
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
* 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
* 				Marian Pohling, Lukas Twardon
*/
package de.unibi.airobots.bonsai.strategies.exploration;

import de.unibi.airobots.bonsai.util.helper.annotation.MapRegion;
import de.unibi.airobots.btl.data.NavigationGoalData;
import de.unibi.airobots.btl.data.PositionData;

/**
 * This interface defines strategies, that produce navigation goals in order to
 * explore the current environment.
 * 
 * @author lziegler
 */
public interface ExplorationStrategy {

    /**
     * Calculates a new navigation goal depending on the robot's current
     * position.
     * 
     * @param currentPosition
     *            The current position of the robot.
     * @throws NoGoalFoundException
     *             Is thrown if no further goal could be generated.
     * @return A reasonable new goal.
     */
    NavigationGoalData getNextGoal(PositionData currentPosition)
            throws NoGoalFoundException;

    /**
     * Calculates a new navigation goal depending on the robot's current
     * position.
     * 
     * @param currentPosition
     *            The current position of the robot.
     * @param region
     *            The region, that should be explored.
     * @throws NoGoalFoundException
     *             Is thrown if no further goal could be generated.
     * @return A reasonable new goal.
     */
    NavigationGoalData getNextGoal(PositionData currentPosition,
            MapRegion region) throws NoGoalFoundException;

    /**
     * This exception should be thrown, if no further navigation goal could be
     * generated.
     * 
     * @author lziegler
     */
    @SuppressWarnings("serial")
    public class NoGoalFoundException extends Exception {

        /**
         * Creates a new exception.
         */
        public NoGoalFoundException() {
            super();
        }

        /**
         * Creates a new exception.
         * 
         * @param message
         *            A user specified error message.
         */
        public NoGoalFoundException(String message) {
            super(message);
        }
    }
}
