/*
* This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/bonsai
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
* 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
* 				Marian Pohling, Lukas Twardon
*/
package de.unibi.airobots.bonsai.strategies.exploration;

import java.io.IOException;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.core.SensorListener;
import de.unibi.airobots.bonsai.util.CoordinateSystemConverter;
import de.unibi.airobots.bonsai.util.helper.annotation.MapRegion;
import de.unibi.airobots.btl.data.LaserData;
import de.unibi.airobots.btl.data.NavigationGoalData;
import de.unibi.airobots.btl.data.PositionData;

/**
 * This class implements an exploration strategy, that searches for the greatest
 * open space region in the laser range.
 * 
 * @author nkoester, sebschne, lziegler
 */
public class OpenSpaceExploration implements ExplorationStrategy,
        SensorListener<LaserData> {

    private static final int INITIAL_TIMEOUT = 100;

    private LaserData currentLaser;
    private Object laserLock = new Object();

    private Logger logger = Logger.getLogger(this.getClass());

    /**
     * Constructs a new instance of this strategy.
     * 
     * @param laserSensor
     *            A sensor receiving laser data.
     */
    public OpenSpaceExploration(Sensor<LaserData> laserSensor) {
        try {
            currentLaser = laserSensor.readLast(INITIAL_TIMEOUT);
        } catch (IOException e) {
            logger.warn("IO error while fetching initial slam map.");
            logger.debug(e);
        } catch (InterruptedException e) {
            logger.warn("Interrupted while fetching initial slam map.");
            logger.debug(e);
        }
        laserSensor.addSensorListener(this);
    }

    @Override
    public void newDataAvailable(LaserData newData) {
        synchronized (laserLock) {
            currentLaser = newData;
        }
    }

    @Override
    public NavigationGoalData getNextGoal(PositionData currentPosition) {

        LaserData laserData = null;
        synchronized (laserLock) {
            laserData = currentLaser;
        }
        if (laserData == null) {
            NavigationGoalData goal = new NavigationGoalData(currentPosition);
            goal.setCoordinateTolerance(0.2);
            goal.setYawTolerance(2 * Math.PI);
            return goal;
        }

        double[] newLaserDataValues = laserData.getScanValues();

        // use segments of 60 laser samples, resulting in 6 segments, each 30
        // degrees wide
        int n = 60;
        double bestLaserAngle = 0.0;
        double bestLaserDist = 0.0;

        for (int i = 0; i < newLaserDataValues.length - n; i += n) {
            double minDist = 8;
            for (int j = i; j < i + n - 1; j++) {
                minDist = Math.min(minDist, newLaserDataValues[j]);
            }
            // TODO This is some arbitrary value to decide, when to turn
            if (minDist > 0.7) {
                if (minDist > bestLaserDist) {
                    bestLaserAngle = LaserData.getAngleValue(i + n / 2);
                    bestLaserDist = minDist;
                }
            }

        }
        NavigationGoalData newNavigationGoalData =
                CoordinateSystemConverter.polar2NavigationGoalData(
                        currentPosition, bestLaserAngle, bestLaserDist * 0.4);

        if (bestLaserDist < 0.3) {
            // if robot should turn, then the tolerance of the angle value must
            // be small
            newNavigationGoalData =
                    CoordinateSystemConverter.polar2NavigationGoalData(
                            currentPosition, Math.PI, 0.0f);
            newNavigationGoalData.setYawTolerance(10 * Math.PI / 180);
            logger.info("mainly turning");
        } else {
            // otherwise we don't care about orientation after motion
            newNavigationGoalData.setYawTolerance(2 * Math.PI);
        }

        logger.info("odometry=" + currentPosition);
        logger.info("bestLaserDist=" + bestLaserDist + " bestLaserAngle="
                + bestLaserAngle + " ("
                + (((bestLaserAngle - 180) / 2) * Math.PI / 180) + ")");

        logger.info("NavigationGoal is x=" + newNavigationGoalData.getX()
                + " y=" + newNavigationGoalData.getY() + " yaw="
                + newNavigationGoalData.getYaw());

        return newNavigationGoalData;
    }

    @Override
    public NavigationGoalData getNextGoal(PositionData currentPosition,
            MapRegion region) {

        // TODO implement this!!
        return getNextGoal(currentPosition);
    }

}
