/*
* This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/bonsai
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
* 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
* 				Marian Pohling, Lukas Twardon
*/
package de.unibi.airobots.bonsai.util.slam.gui;

import static java.lang.Math.cos;
import static java.lang.Math.sin;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.core.BonsaiManager;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.core.configuration.ConfigurationParser;
import de.unibi.airobots.bonsai.core.configuration.ConfiguredActuator;
import de.unibi.airobots.bonsai.core.configuration.ConfiguredCoreObjectFactory;
import de.unibi.airobots.bonsai.core.configuration.ConfiguredSensor;
import de.unibi.airobots.bonsai.util.Pair;
import de.unibi.airobots.bonsai.util.slam.SlamTools;
import de.unibi.airobots.bonsai.xcf.XcfFactory;
import de.unibi.airobots.bonsai.xcf.actuators.XcfNavigationActuator;
import de.unibi.airobots.bonsai.xcf.configuration.ConfiguredXcfBinarySlamSensor;
import de.unibi.airobots.bonsai.xcf.configuration.ConfiguredXcfBtlPublisherSensor;
import de.unibi.airobots.bonsai.xcf.configuration.ConfiguredXcfNavigationActuator;
import de.unibi.airobots.bonsai.xcf.sensors.XcfBinarySlamSensor;
import de.unibi.airobots.bonsai.xcf.sensors.XcfBtlPublisherSensor;
import de.unibi.airobots.btl.data.BinarySlamMap;
import de.unibi.airobots.btl.data.GlobalPlan;
import de.unibi.airobots.btl.data.NavigationGoalData;
import de.unibi.airobots.btl.data.PositionData;

/**
 * A panel for displaying a (SLAM) map and overlaying things on top of it. You
 * can also register listeners for clicks into the world.
 * 
 * @author dklotz
 */
public class MapPanel extends JPanel {

    /** To make Serializable happy... */
    private static final long serialVersionUID = 3347424196567633051L;

    /** For logging output. */
    private static final Logger LOGGER = Logger.getLogger(MapPanel.class);

    /** For getting the map. */
    private Sensor<BinarySlamMap> mapSensor;

    /** For getting the position. */
    private Sensor<PositionData> positionSensor;
    // TODO: Use this insted of SlamSensor, fix Sensor.java first?

    /** For notifying people of clicks into the world map. */
    private final List<WorldClickListener> listeners =
            new ArrayList<WorldClickListener>();

    /** For getting the current navigation goal. */
    private final NavigationActuator navigation;

    private final Sensor<GlobalPlan> planSensor;

    /** The last image version of the slam map. */
    private BufferedImage slamImage;
    /** The latest slam data. */
    private BinarySlamMap slamMap;
    /** the last interpolated Pos. */
    private PositionData interpolPos;
    /** The latest navigation goal. */
    private NavigationGoalData goal;

    private GlobalPlan planData;

    /** Used for periodically updating the map data. */
    private final ScheduledExecutorService executor =
            Executors.newSingleThreadScheduledExecutor();

    /** Used for transforming the image. */
    private AffineTransform imageTransform;
    /** Used for transforming all the things drawn on top of the image. */
    private AffineTransform pointTransform;

    // TODO: Rotate the map (world axis aligned to window)?
    // TODO: Find some elegant method for the zoom level?

    /**
     * Creates a new map panel.
     * 
     * @param mapSensor
     *            Used for getting the current SLAM map, should not be null.
     * @param navigation
     *            Used for getting the current goal of the navigation algorithm.
     * @param positionSensor
     *            Used for getting the current robot position.
     * @param planSensor
     *            Used for getting the current global plan.
     * @param updateTime
     *            How fast (in ms) should we try to update the SLAM map?
     */
    public MapPanel(final Sensor<BinarySlamMap> mapSensor,
            final NavigationActuator navigation,
            final Sensor<PositionData> positionSensor,
            final Sensor<GlobalPlan> planSensor, long updateTime) {
        super(new BorderLayout());

        this.setPreferredSize(new Dimension(570, 500));

        this.mapSensor = mapSensor;
        this.positionSensor = positionSensor;
        this.navigation = navigation;
        this.planSensor = planSensor;

        this.addMouseListener(new MouseAdapter() {

            @Override
            public void mouseClicked(MouseEvent e) {
                notifyWorldClickListeners(e);
            }

        });

        executor.scheduleWithFixedDelay(new Runnable() {

            @Override
            public void run() {
                updateMap();
            }
        }, 200, updateTime, TimeUnit.MILLISECONDS);

    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#finalize()
     */
    @Override
    // CHECKSTYLE:OFF
    protected void finalize() throws Throwable {
        // CHECKSTYLE:ON
        super.finalize();
        executor.shutdownNow();
    }

    /**
     * Tries to update the SLAM map from the sensor.
     */
    private synchronized void updateMap() {
        try {
            interpolPos = positionSensor.readLast();
            slamMap = mapSensor.readLast(0);
            if (slamMap != null) {
                slamImage = SlamTools.slamMapToImage(slamMap);

                // TODO: Make scaling respect the maps aspect ratio?
                imageTransform =
                        new AffineTransform(getWidth()
                                / (double) slamMap.getWidth(), 0, 0,
                                getHeight() / (double) slamMap.getHeight(), 0,
                                0);
                pointTransform = new AffineTransform(imageTransform);
                pointTransform.scale(1, -1);
                pointTransform.translate(0, -(slamMap.getHeight() - 1));
            }
            try {
                goal = navigation.getCurrentGoal();
            } catch (IOException e) {
                goal = null;
            }

            try {
                LOGGER.debug("Updating global plan");
                if (planSensor.hasNext()) {
                    planData = planSensor.readLast(100);
                    LOGGER.debug("Got new global plan");
                }
            } catch (IOException e) {
                LOGGER.warn("Error updating global plan", e);
            }

            SwingUtilities.invokeLater(new Runnable() {

                @Override
                public void run() {
                    repaint();
                }
            });
            // CHECKSTYLE:OFF we really want to catch everything here
        } catch (Throwable e) {
            // CHECKSTYLE:ON
            LOGGER.warn("Throwable caught while updating map panel.", e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.swing.JComponent#paint(java.awt.Graphics)
     */
    @Override
    public void paint(Graphics g) {

        Graphics2D g2 = (Graphics2D) g;

        g2.clearRect(0, 0, getWidth(), getHeight());

        drawMap(g2);
    }

    /**
     * Draws the SLAM map.
     * 
     * @param g2
     *            The {@link Graphics2D} object used to draw.
     */
    private synchronized void drawMap(Graphics2D g2) {
        if (slamImage != null) {
            // Draw transformed image
            g2.drawImage(slamImage, imageTransform, this);

            // Draw green circle at slam origin
            g2.setColor(Color.GREEN);
            double[] slamOrigin =
                    new double[] { slamMap.getOriginX(), slamMap.getOriginY() };
            pointTransform.transform(slamOrigin, 0, slamOrigin, 0, 1);
            drawPose(g2, (int) slamOrigin[0], (int) slamOrigin[1], 0.0, true,
                    "(0,0)");

            // Draw red circle at current slam position
            int slamPosX = slamMap.x2idx(slamMap.getX());
            int slamPosY = slamMap.y2idx(slamMap.getY());
            g2.setColor(Color.RED);
            double[] slamPos = new double[] { slamPosX, slamPosY };
            pointTransform.transform(slamPos, 0, slamPos, 0, 1);
            drawPose(g2, (int) slamPos[0], (int) slamPos[1], slamMap.getYaw(),
                    true, String.format("(%.3f,%.3f)", slamMap.getX(), slamMap
                            .getY()));

            // Draw orange circle at current interpolated position
            int interpolPosX = slamMap.x2idx(interpolPos.getX());
            int interpolPosY = slamMap.y2idx(interpolPos.getY());
            g2.setColor(Color.ORANGE);
            double[] interpolPosVec =
                    new double[] { interpolPosX, interpolPosY };
            pointTransform.transform(interpolPosVec, 0, interpolPosVec, 0, 1);
            drawPose(g2, (int) interpolPosVec[0], (int) interpolPosVec[1],
                    interpolPos.getYaw(), true, String.format("(%.3f,%.3f)",
                            interpolPos.getX(), interpolPos.getY()));

            if (goal != null) {
                // Draw blue circle at current navigation goal position
                g2.setColor(Color.BLUE);
                double[] navGoal =
                        new double[] { slamMap.x2idx(goal.getX()),
                                slamMap.y2idx(goal.getY()) };
                pointTransform.transform(navGoal, 0, navGoal, 0, 1);
                drawPose(g2, (int) navGoal[0], (int) navGoal[1], goal.getYaw(),
                        true, String.format("(%.3f,%.3f)", goal.getX(), goal
                                .getY()));
            }

            // draw current global plan
            if (planData != null) {
                LOGGER.info("Drawing plan of size " + planData.size());
                g2.setColor(Color.GREEN);
                List<Integer> xs = new LinkedList<Integer>();
                List<Integer> ys = new LinkedList<Integer>();
                for (NavigationGoalData wp : planData) {
                    double[] pointCoords =
                            new double[] { slamMap.x2idx(wp.getX()),
                                    slamMap.y2idx(wp.getY()) };
                    pointTransform.transform(pointCoords, 0, pointCoords, 0, 1);
                    xs.add((int) pointCoords[0]);
                    ys.add((int) pointCoords[1]);
                }
                int xArray[] = new int[xs.size()];
                int yArray[] = new int[ys.size()];
                for (int i = 0; i < xs.size(); i++) {
                    xArray[i] = xs.get(i);
                    yArray[i] = ys.get(i);
                }
                g2.drawPolyline(xArray, yArray, xArray.length);
            }
        }
    }

    /**
     * Draw a pose (position and optionally orientation) in the map.
     * 
     * @param g2
     *            The graphics object used for drawing.
     * @param x
     *            The x coordinate for the pose (in image coordinates!).
     * @param y
     *            The y coordinate for the pose (in image coordinates!).
     * @param yaw
     *            The orientation angle (0.0 points to the right).
     * @param drawOrientation
     *            If we should draw the orientation.
     * @param description
     *            Optional description displayed at the pose.
     */
    private void drawPose(Graphics2D g2, int x, int y, double yaw,
            boolean drawOrientation, String description) {
        // Draw a circle centered on the position
        int circleRadius = 5;
        g2.drawOval(x - circleRadius, y - circleRadius, (circleRadius * 2) + 1,
                (circleRadius * 2) + 1);

        // Optionally draw a line from the position in the direction of the
        // orientation
        if (drawOrientation) {
            float lineLength = circleRadius + 3;
            g2.drawLine(x, y, x + (int) (cos(yaw) * lineLength), y
                    - (int) (sin(yaw) * lineLength));
        }

        // Optionally draw a description to the top right of the circle
        if (!(description == null || "".equals(description.trim()))) {
            g2.drawString(description, x + circleRadius, y - circleRadius);
        }
    }

    /**
     * Converts pixel coordinates in this panel to coordinates in the slam maps
     * occupancy grid.
     * 
     * @param panelCoordinates
     *            Pixel coordinates in this panel, must not be null.
     * @return Grid indices in the occupancy grid.
     */
    private Pair<Integer> panelCoordinates2Grid(Pair<Integer> panelCoordinates) {
        try {
            // We just reverse the transformation from grid into pixel
            // coordinates and use that to transform the coordinates in the
            // other direction.
            AffineTransform reverse = pointTransform.createInverse();
            double[] point =
                    new double[] { panelCoordinates.getFirst(),
                            panelCoordinates.getSecond() };
            reverse.transform(point, 0, point, 0, 1);
            return new Pair<Integer>((int) point[0], (int) point[1]);
        } catch (NoninvertibleTransformException e) {
            // This should normally never happen!
            LOGGER.warn("Could not invert SLAM map transform.", e);
            throw new IllegalStateException("Could not invert transformation "
                    + "that should be invertible?!");
        }
    }

    /**
     * Notifies all listeners of a click on the map.
     * 
     * @param panelCoordinates
     *            Click coordinates, in pixels on this panel!
     */
    private void notifyWorldClickListeners(MouseEvent mouseEvent) {
        Pair<Integer> panelCoordinates =
                new Pair<Integer>(mouseEvent.getX(), mouseEvent.getY());
        Pair<Integer> gridCoordinates = panelCoordinates2Grid(panelCoordinates);
        Pair<Double> worldCoordinates;

        // This is synchronized so both calls are guaranteed to refer to the
        // same slam map
        synchronized (this) {
            double worldX = slamMap.idx2x(gridCoordinates.getFirst());
            double worldY = slamMap.idx2y(gridCoordinates.getSecond());

            worldCoordinates = new Pair<Double>(worldX, worldY);
        }

        // Notify all listeners
        List<WorldClickListener> listenersCopy =
                new ArrayList<WorldClickListener>(listeners);
        for (WorldClickListener listener : listenersCopy) {
            listener.worldClicked(worldCoordinates, mouseEvent);
        }
    }

    /**
     * Adds a new listeners for clicks into the world.
     * 
     * @param listener
     *            The new listener, should not be null.
     */
    public void addWorldClickListener(WorldClickListener listener) {
        this.listeners.add(listener);
    }

    /**
     * Removes a listener for clicks into the world.
     * 
     * @param listener
     *            The listener to be removed, should not be null.
     * @return <code>true</code>, if this listener was really registered.
     */
    public boolean removeWorldClickListener(WorldClickListener listener) {
        return this.listeners.remove(listener);
    }

    /**
     * Just for testing purposes.
     * 
     * @param args
     *            Ignored.
     */
    public static void main(String[] args) {

        BasicConfigurator.configure();

        LOGGER.debug("Got manager, trying to get sensor...");

        BonsaiManager manager = BonsaiManager.getInstance();
        manager.setParser(new ConfigurationParser() {

            @Override
            public Set<ConfiguredActuator> getConfiguredActuators()
                    throws IllegalStateException {

                ConfiguredActuator navigationAct = new ConfiguredActuator();
                navigationAct.setActuatorClassName(XcfNavigationActuator.class
                        .getName());
                navigationAct.setFactoryClassName(XcfFactory.class.getName());
                navigationAct.setInterfaceClassName(NavigationActuator.class
                        .getName());
                navigationAct.setKey("NavigationActuator");
                Map<String, String> navActOptions =
                        new HashMap<String, String>();
                navActOptions.put(
                        ConfiguredXcfNavigationActuator.OPTION_SERVER_NAME,
                        "Sunflower");
                navigationAct.setOptions(navActOptions);

                HashSet<ConfiguredActuator> set =
                        new HashSet<ConfiguredActuator>();
                set.add(navigationAct);
                return set;

            }

            @Override
            public Set<ConfiguredCoreObjectFactory> getConfiguredFactories()
                    throws IllegalStateException {

                ConfiguredCoreObjectFactory xcfFactory =
                        new ConfiguredCoreObjectFactory();
                xcfFactory.setFactoryClassName(XcfFactory.class.getName());
                xcfFactory.setFactroyOptions(new HashMap<String, String>());

                HashSet<ConfiguredCoreObjectFactory> set =
                        new HashSet<ConfiguredCoreObjectFactory>();
                set.add(xcfFactory);
                return set;

            }

            @Override
            public Set<ConfiguredSensor> getConfiguredSensors()
                    throws IllegalStateException {

                ConfiguredSensor slamSensor = new ConfiguredSensor();
                slamSensor.setDataType(BinarySlamMap.class.getName());
                slamSensor.setFactoryClassName(XcfFactory.class.getName());
                slamSensor.setKey("SlamSensor");
                slamSensor.setSensorClassName(XcfBinarySlamSensor.class
                        .getName());
                Map<String, String> slamSensorOptions =
                        new HashMap<String, String>();
                slamSensorOptions.put(
                        ConfiguredXcfBinarySlamSensor.OPTION_PUBLISHER_NAME,
                        "SlamMap");
                slamSensor.setOptions(slamSensorOptions);

                ConfiguredSensor globalPlanSensor = new ConfiguredSensor();
                globalPlanSensor.setDataType(GlobalPlan.class.getName());
                globalPlanSensor
                        .setFactoryClassName(XcfFactory.class.getName());
                globalPlanSensor.setKey("GlobalPlanSensor");
                globalPlanSensor.setSensorClassName(XcfBtlPublisherSensor.class
                        .getName());
                Map<String, String> globalPlanSensorOptions =
                        new HashMap<String, String>();
                globalPlanSensorOptions.put(
                        ConfiguredXcfBtlPublisherSensor.OPTION_PUBLISHER_NAME,
                        "GlobalPlan");
                globalPlanSensor.setOptions(globalPlanSensorOptions);

                HashSet<ConfiguredSensor> set = new HashSet<ConfiguredSensor>();
                set.add(slamSensor);
                set.add(globalPlanSensor);
                return set;

            }

            @Override
            public void parseFile(URI configurationUrl) throws IOException,
                    ParseException, IllegalStateException {
                // ignore this
            }

        });
        manager.configure("foo");

        Sensor<BinarySlamMap> slamSensor =
                manager.createSensor("SlamSensor", BinarySlamMap.class);
        Sensor<PositionData> posSensor =
                manager.createSensor("SlamSensor", PositionData.class);
        Sensor<GlobalPlan> gpsSensor =
                manager.createSensor("GlobalPlanSensor", GlobalPlan.class);

        final NavigationActuator actuator =
                manager.createActuator("NavigationActuator",
                        NavigationActuator.class);

        LOGGER.debug("Got sensor & actuator, starting GUI...");

        MapPanel panel =
                SlamTools.showMapPanelWindow(slamSensor, posSensor, actuator,
                        gpsSensor, false);

        JFrame frame = (JFrame) SwingUtilities.getRoot(panel);
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

    }

    /**
     * A listeners for clicks on the world map.
     * 
     * @author dklotz
     */
    public static interface WorldClickListener {

        /**
         * Called whenever the user has clicked on a location in the global map.
         * 
         * @param worldCoordinates
         *            Global world position the user has clicked on, never null.
         * @param mouseEvent
         *            For getting things like the click count or which button
         *            was clicked, never null.
         */
        void worldClicked(Pair<Double> worldCoordinates, MouseEvent mouseEvent);
    }
}
