/*
* This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
*
* Copyright(c) Frederic Siepmann
* http://opensource.cit-ec.de/projects/bonsai
*
* This file may be licensed under the terms of of the
* GNU Lesser General Public License Version 3 (the ``LGPL''),
* or (at your option) any later version.
*
* Software distributed under the License is distributed
* on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
* express or implied. See the LGPL for the specific language
* governing rights and limitations.
*
* You should have received a copy of the LGPL along with this
* program. If not, go to http://www.gnu.org/licenses/lgpl.html
* or write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
* The development of this software was supported by the
* Excellence Cluster EXC 277 Cognitive Interaction Technology.
* The Excellence Cluster EXC 277 is a grant of the Deutsche
* Forschungsgemeinschaft (DFG) in the context of the German
* Excellence Initiative.
*
* Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
* 				Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
* 				Marian Pohling, Lukas Twardon
*/
package de.unibi.airobots.bonsai.xcf;

import java.lang.reflect.InvocationTargetException;

import org.apache.log4j.Logger;

import Ice.ConnectFailedException;

import net.sf.xcf.MethodNotFoundException;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.UserException;

/**
 * A wrapper for {@link RemoteServer}s that passes all connection exceptions
 * (and all other exceptions, too) to the caller. The difference to a normal
 * {@link RemoteServer} is that it notifies an instance of
 * {@link RemoteServerReconnectableManager} before passing the connection
 * exception to the caller.
 * 
 * @author jwienke
 */
public class ConnectionErrorNotifyingRemoteServer extends RemoteServerWrapper {

    private Logger logger = Logger.getLogger(getClass());

    /**
     * Constructor.
     * 
     * @param wrappedServer
     *            wrapped remote server
     * @param manager
     *            {@link RemoteServerReconnectableManager} that manages this
     *            wrapper
     */
    ConnectionErrorNotifyingRemoteServer(RemoteServer wrappedServer,
            RemoteServerReconnectableManager manager) {
        super(wrappedServer, manager);
    }

    /**
     * Notifies the manager on connections errors and then passes the exception
     * back to the client of this class.
     * 
     * {@inheritDoc}
     */
    @Override
    // CHECKSTYLE:OFF we really want to throw everything that is possible here
    // because XCF already does this.
    protected Object reflectiveCallMethod(RemoteServerMethod method,
            Object... args) throws RuntimeException, UserException,
            MethodNotFoundException {
        // CHECKSTYLE:ON

        try {
            return invokeWrappedServerMethod(method, args);
        } catch (InvocationTargetException e) {

            logger.debug("Calling remote server method " + method + " failed",
                    e.getCause());

            if (e.getCause() instanceof ConnectFailedException) {
                markWrappedServerAsDisconnected();
                throw (ConnectFailedException) e.getCause();
            } else if (e.getCause() instanceof UserException) {
                throw (UserException) e.getCause();
            } else if (e.getCause() instanceof MethodNotFoundException) {
                throw (MethodNotFoundException) e.getCause();
            } else {
                logger.fatal("Unexpected exception in "
                        + "wrapped remote server.", e);
                assert false;
                throw new RuntimeException(e.getCause());
            }

        }

    }

}
